#!/usr/bin/python3
# -*- coding: utf-8 -*-

import dbus
import dbus.service
import dbus.mainloop.glib
import os
import time
import datetime
import threading
import shutil
from kylindbus.apt_handler import AppActions, AptHandler, WorkitemError
from utils.enums import INTERFACE, INTERFACES, AUTOSTART, CRONDAILY
from kylindbus.policykit_service import PolicyKitService
import aptsources.sourceslist
import subprocess
from PyQt5.QtCore import pyqtSignal

CRONPATH = "/var/spool/anacron/"
SOURCESPATH = "/etc/apt/sources.list.d"
CODEHOME = "/usr/share/kylin-update-manager/"


class WorkItem:
     def __init__(self, pkgname, action, kwargs):
        self.pkgname = pkgname
        self.action = action
        self.kwargs = kwargs

class WorkThread(threading.Thread):

    def __init__(self, dbusDaemon):
        threading.Thread.__init__(self)
        self.dbusDaemon = dbusDaemon
        print("KUM DBUS START ", type(self.dbusDaemon))

    #
    # 函数:线程执行函数
    #
    def run(self):
        while(True):

            if len(self.dbusDaemon.worklist) == 0:
                time.sleep(0.2)
                continue

            self.dbusDaemon.mutex.acquire()
            item = self.dbusDaemon.worklist.pop(0)
            self.dbusDaemon.mutex.release()
            try:
                func = getattr(self.dbusDaemon.aptHandler, item.action)
                if func is None:
                    print("Error action: ", item)
                res = func(item.pkgname,item.kwargs)
                if res is False:
                    print("Action exec failed...")
            except WorkitemError as e:
                # 如果是包列表，此时强制装换为str，后面显示str将出现异常，故而暂不显示包列表
                kwarg = {"apt_appname": str(item.pkgname),
                        "apt_percent": str(-e.errornum),
                        "action": str(item.action),
                        }
                self.dbusDaemon.kum_apt_signal("apt_error", kwarg)
            except:
                kwarg = {"apt_appname": str(item.pkgname),
                        "apt_percent": '-24',
                        "action": str(item.action),
                        }
                self.dbusDaemon.kum_apt_signal("apt_error", kwarg)
            time.sleep(0.3)


class Daemon(PolicyKitService):
    down_load_pid = ""
    def __init__ (self, bus, mainloop):
        self.aptHandler = AptHandler(self)
        self.worklist = []
        self.cancel_name_list = []
        self.mutex = threading.RLock()
        self.cancelmutex = threading.RLock()
        self.worker_thread = WorkThread(self)
        self.worker_thread.setDaemon(True)
        self.worker_thread.start()


        self.bus_name = dbus.service.BusName(INTERFACE, bus=bus)
        PolicyKitService.__init__(self, self.bus_name, INTERFACES)
        self.mainloop = mainloop

    #
    # 函数:停止dbus
    #
    def stop(self):
        """ stop the dbus controller and remove from the bus """
        self.remove_from_connection()

    #
    # 函数:新增任务
    # item：任务
    #
    def add_worker_item(self, item):
        print("####add_worker_item:", item)
        self.mutex.acquire()
        self.worklist.append(item)
        self.mutex.release()

    #
    # 函数:删除任务
    # pkgname：任务名
    #
    def del_worker_item_by_name(self, pkgname):
        print("####del_worker_item_by_name:", pkgname)
        exist = False
        self.mutex.acquire()
        for item in self.worklist:
            if item.pkgname == pkgname:
                exist = True
                break
        if exist is True:
            self.worklist.remove(pkgname)
        self.mutex.release()

        self.cancelmutex.acquire()
        self.cancel_name_list.append(pkgname)
        self.cancelmutex.release()
        print("####del_worker_item_by_name finished!")

    #
    # 函数:检查取消的任务
    #
    def check_cancel_worker_item(self, pkgname):
#        print "####check_cancel_worker_item:",pkgname
        cancel = False
        self.cancelmutex.acquire()
#        print "check_cancel_worker_item:",len(self.cancel_name_list)
        for item in self.cancel_name_list:
            if item == pkgname:
                cancel = True
                break
        if cancel is True:
            self.cancel_name_list.remove(pkgname)
        self.cancelmutex.release()

#        print "####check_cancel_worker_item finished!:",cancel
        return cancel

    #
    # 函数:退出dbus
    #
    @dbus.service.method(INTERFACE, in_signature='', out_signature='')
    def exit(self):
        self.mainloop.quit()

    #
    # 函数:初始化cache
    #
    @dbus.service.method(INTERFACE, in_signature='', out_signature='')
    def init_cache(self):
        self.aptHandler.init_cache()

    #
    # 函数:安装和升级
    # install package sa:kum_fetch_signal() and kum_apt_signal()
    #
    @dbus.service.method(INTERFACE, in_signature='s', out_signature='b', sender_keyword='sender')
    def install_and_upgrade(self, pkgName, sender=None):
        print("####install_and_upgrade: ", pkgName)

        item = WorkItem(pkgName, AppActions.INS_AND_UP, None)

        self.add_worker_item(item)

        print("####install_and_upgrade return")
        return True

    #
    # 函数：安装
    # install package sa:kum_fetch_signal() and kum_apt_signal()
    #
    @dbus.service.method(INTERFACE, in_signature='as', out_signature='b', sender_keyword='sender')
    def install(self, pkgNames, sender=None):
        print("####install: ", pkgNames)

        item = WorkItem(pkgNames, AppActions.INSTALL, None)

        self.add_worker_item(item)

        print("####install return")
        return True

    #
    # 函数：单个安装
    #
    @dbus.service.method(INTERFACE, in_signature='s', out_signature='b', sender_keyword='sender')
    def insone(self, pkgName, sender=None):
        print("####insone: ", pkgName)

        item = WorkItem(pkgName, AppActions.INSTALL_ONE, None)

        self.add_worker_item(item)

        print("####insone return")
        return True

    #
    # 函数：升级
    # update package sa:kum_fetch_signal() and kum_apt_signal()
    #
    @dbus.service.method(INTERFACE, in_signature='as', out_signature='b', sender_keyword='sender')
    def upgrade(self, pkgNames, sender=None):
        print("####upgrade: ", pkgNames)

        item = WorkItem(pkgNames, AppActions.UPGRADE, None)

        self.add_worker_item(item)

        print("####upgrade return")
        return True

    #
    # 函数：单个升级
    #
    @dbus.service.method(INTERFACE, in_signature='s', out_signature='b', sender_keyword='sender')
    def upone(self, pkgName, sender=None):
        print("####upone: ", pkgName)

        item = WorkItem(pkgName, AppActions.UPGRADE_ONE, None)

        self.add_worker_item(item)

        print("####upone return")
        return True

    #
    # 函数：取消
    #
    @dbus.service.method(INTERFACE, in_signature='s', out_signature='b', sender_keyword='sender')
    def cancel(self, pkgName, sender=None):
        print("####cancel: ", pkgName)

        self.del_worker_item_by_name(pkgName)

        print("####cancel return")
        return True

    #
    # 函数：更新软件源
    # apt-get update sa:kum_fetch_signal()
    #
    @dbus.service.method(INTERFACE, in_signature='b', out_signature='b', sender_keyword='sender')
    def update(self, quiet, sender=None):
        print("####update: ")

        kwargs = {"quiet":str(quiet),
                  }

        item = WorkItem("#update", AppActions.UPDATE, kwargs)

        self.add_worker_item(item)

        print("####update return")
        return True

    #
    # 函数：通过shell更新软件源
    #
    @dbus.service.method(INTERFACE, in_signature='b', out_signature='b', sender_keyword='sender')
    def update_by_shell(self, quiet, sender=None):
        print("####update by shell: ")

        kwargs = {"quiet":str(quiet),
                  }

        item = WorkItem("#update-shell", AppActions.UPDATESHELL, kwargs)

        self.add_worker_item(item)

        print("####update by shell return")
        return True

    #
    # 函数：解决冲突
    #
    @dbus.service.method(INTERFACE, in_signature='b', out_signature='b', sender_keyword='sender')
    def configure_dpkg_by_shell(self, quiet, sender=None):
        print("####configure dpkg by shell: ")

        kwargs = {"quiet":str(quiet),
                  }

        item = WorkItem("#configure-shell", AppActions.CONFIGURESHELL, kwargs)

        self.add_worker_item(item)

        print("####configure dpkg by shell return")
        return True

    #
    # 函数：通过shell解决冲突
    #
    @dbus.service.method(INTERFACE, in_signature='b', out_signature='b', sender_keyword='sender')
    def fix_conffile_by_shell(self, quiet, sender=None):
        print("####fix conffile by shell: ")

        kwargs = {"quiet":str(quiet),
                  }

        item = WorkItem("#fix-conffile-shell", AppActions.FIXCONFFILE, kwargs)

        self.add_worker_item(item)

        print("####fix conffile by shell return")
        return True

    #
    # 函数：开机自启动开启
    #
    @dbus.service.method(INTERFACE, in_signature='', out_signature='b', sender_keyword='sender')
    def autostart_open(self, sender=None):
        try:
            cmd = "cp " + CODEHOME + "kylin-update-manager-autostart.desktop  " + AUTOSTART
            os.popen(cmd)
            return True
        except:
            return False

    #
    # 函数：开机自启动关闭
    #
    @dbus.service.method(INTERFACE, in_signature='', out_signature='b', sender_keyword='sender')
    def autostart_close(self, sender=None):
        try:
            cmd = "rm " + AUTOSTART + "kylin-update-manager-autostart.desktop"
            os.popen(cmd)
            return True
        except:
            return False

    #
    # 函数：每日定时启动开启
    #
    @dbus.service.method(INTERFACE, in_signature='', out_signature='b', sender_keyword='sender')
    def dailystart_open(self, sender=None):
        try:
            cmd = "cp " + CODEHOME + "0kylin-update-manager-daily  " + CRONDAILY
            os.popen(cmd)
            return True
        except:
            return False

    #
    # 函数：每日定时启动关闭
    #
    @dbus.service.method(INTERFACE, in_signature='', out_signature='b', sender_keyword='sender')
    def dailystart_close(self, sender=None):
        try:
            cmd = "rm " + CRONDAILY + "0kylin-update-manager-daily"
            os.popen(cmd)
            return True
        except:
            return False

    #-------------------------------------signals----------------------------------------

    @dbus.service.signal(INTERFACE, signature='sa{ss}')
    def software_auth_signal(self, type, msg):
        pass

    #
    # 函数：下载消息信号
    # package download status signal
    '''parm mean
        type:
            start:start download
            stop:all work is finish
            done:all items download finished
            fail:download failed
            fetch:one item download finished
            pulse:download status, this msg given a string like dict
        msg:
            a message of type, sometimes is None
    '''
    @dbus.service.signal(INTERFACE, signature='sa{ss}')
    def kum_fetch_signal(self, type_msg, msg):
        pass

    #
    # 函数：执行消息信号
    # package install/update/remove signal
    '''parm mean
        type:
            start:start work
            stop:work finish
            error:got a error
            pulse:work status, this msg given a string like dict
        msg:
            a message of type, sometimes is None
    '''
    @dbus.service.signal(INTERFACE, signature='sa{ss}')
    def kum_apt_signal(self, type_msg, msg):
        pass

    # test func
    @dbus.service.method(INTERFACE, in_signature='i', out_signature='i')
    def test_return_value(self, arg):
        return 2

    # @dbus.service.method(INTERFACE, in_signature='sssss', out_signature='b')
    # def change_source_list_to_default(self, service_key, pwd, current_user, os_codename, available_source):
    #     if os.path.exists(SOURCESPATH):
    #         for item in os.listdir(SOURCESPATH):
    #             targetFile = os.path.join(SOURCESPATH, item)
    #             if os.path.isfile(targetFile):
    #                 os.remove(targetFile)
    #
    #     cdrom_addr = "file:///media/" + current_user + "/Kylin-4.0"
    #     ginkgo_addr = "http://" + service_key + ":" + pwd + "@archive.kylinos.cn/"
    #     juniper_addr = "http://" + service_key + ":" + pwd + "@archive.kylinos.cn/kylin/juniper"
    #
    #     if available_source == "no_source":
    #         fp = open('/etc/apt/sources.list', 'w+')
    #         fp.tell()
    #F         if os_codename == "ginkgo":
    #             fp.write("deb " + cdrom_addr + " ginkgo main restricted multiverse universe\n")
    #             fp.write("deb " + ginkgo_addr + " ginkgo main restricted multiverse universe\n")
    #         if os_codename == "juniper":
    #             fp.write("deb " + cdrom_addr + " juniper main restricted multiverse universe\n")
    #             fp.write("deb " + juniper_addr + " juniper main restricted multiverse universe\n")
    #         fp.close()
    #
    #     if available_source == "dvd_source":
    #         fp = open('/etc/apt/sources.list', 'a')
    #         if os_codename == "ginkgo":
    #             fp.write("deb " + ginkgo_addr + " ginkgo main restricted multiverse universe\n")
    #         if os_codename == "juniper":
    #             fp.write("deb " + juniper_addr + " juniper main restricted multiverse universe\n")
    #         fp.close()
    #
    #     if available_source == "ftp_source":
    #         fp = open('/etc/apt/sources.list', 'a')
    #         if os_codename == "ginkgo":
    #             fp.write("deb " + cdrom_addr + " ginkgo main restricted multiverse universe\n")
    #             fp.write("deb " + ginkgo_addr + " ginkgo main restricted multiverse universe\n")
    #         if os_codename == "juniper":
    #             fp.write("deb " + cdrom_addr + " juniper main restricted multiverse universe\n")
    #             fp.write("deb " + juniper_addr + " juniper main restricted multiverse universe\n")
    #         fp.close()
    #
    #     if available_source == "http_source":
    #         fp = open('/etc/apt/sources.list', 'a')
    #         if os_codename == "ginkgo":
    #             fp.write("deb " + cdrom_addr + " ginkgo main restricted multiverse universe\n")
    #         if os_codename == "juniper":
    #             fp.write("deb " + cdrom_addr + " juniper main restricted multiverse universe\n")
    #         fp.close()
    #
    #     return True

    #
    # 函数：恢复默认sourceslist
    #
    @dbus.service.method(INTERFACE, in_signature='sssss', out_signature='b')
    def change_source_list_to_default(self, service_key, pwd, current_user, os_codename, available_source):

        if(os.path.isfile("/etc/apt/sources.list.kumbak")):
            shutil.move("/etc/apt/sources.list.kumbak", "/etc/apt/sources.list")

        # lixiang
        tmppath = '/var/lib/apt/tmp'
        aptpath = '/var/lib/apt/lists'

        if os.path.exists(tmppath) and os.path.isdir(tmppath):
            # remove /var/lib/apt/lists
            for filename in os.listdir(aptpath):
                if filename != 'lock':
                    abspath = os.path.join(aptpath, filename)
                    if os.path.isfile(abspath):
                        os.remove(abspath)

            # copy to /var/lib/apt/lists
            for filename in os.listdir(tmppath):
                if filename != 'lock':
                    abspath = os.path.join(tmppath, filename)
                    if os.path.isfile(abspath):
                        shutil.move(abspath, os.path.join(aptpath, filename))

            # delete /var/lib/apt/tmp
            shutil.rmtree(tmppath)

        return True

    #
    # 函数：恢复升级服务器的sourceslist
    #
    @dbus.service.method(INTERFACE, in_signature='as', out_signature='b')
    def change_source_list_to_kylin_update_server(self, lines):
        # print("chenggong")
        if os.path.exists(SOURCESPATH):
           # print("失败")
           for item in os.listdir(SOURCESPATH):
               targetFile = os.path.join(SOURCESPATH, item)
               if os.path.isfile(targetFile):
                   os.remove(targetFile)

        # wenbo 判断备份文件如果存在，说明上次可能异常中断没进行还原，这次不再进行源备份，防止覆盖上次真正备份的内容
        if(os.path.isfile("/etc/apt/sources.list.kumbak")==False):
            shutil.copy("/etc/apt/sources.list", "/etc/apt/sources.list.kumbak")

            # lixiang
            tmppath = '/var/lib/apt/tmp'
            aptpath = '/var/lib/apt/lists'

            # mkdir /var/lib/apt/tmp
            if os.path.exists(tmppath) and os.path.isdir(tmppath):
                shutil.rmtree(tmppath)
            os.mkdir(tmppath, 0o755)

            for filename in os.listdir(aptpath):
                if filename != 'lock':
                    abspath = os.path.join(aptpath, filename)
                    if os.path.isfile(abspath):
                        cmd = "cp %s %s" %(abspath ,os.path.join(tmppath, filename))
                        #shutil.copy(abspath, os.path.join(tmppath, filename))
                        os.system(cmd)
        fp = open('/etc/apt/sources.list', 'w+')
        fp.tell()
        for line in lines:
           fp.write(line + "\n")
        fp.close()

        return True

    #
    # 函数：wget下载软件包
    #
    @dbus.service.method(INTERFACE, in_signature='sss', out_signature='')
    def copy_file_to_install(self,src_path,appname,dest_debname):
        if(dest_debname != ""):
            dest_path = " /var/cache/apt/archives/" + dest_debname
            cmd = "mv "+src_path +dest_path
        else:
            cmd = "mv "+src_path +" /var/cache/apt/archives/"
        os.system(cmd)
        self.copy_finish(appname)



    #
    #函数:判断拷贝结束
    #
    @dbus.service.signal(INTERFACE,signature="s")
    def copy_finish(self,appname):
        pass

    #
    #函数：暂停下载
    #
    @dbus.service.method(INTERFACE, in_signature='', out_signature='')
    def kill_pid(self):
        os.system("killall -9 wget")

    #
    #函数：点击取消下载
    #
    @dbus.service.method(INTERFACE, in_signature='s', out_signature='')
    def cancel_download_app(self,appname):
        cmd1 = "rm /var/cache/apt/archives/"+appname+"_*"
        os.system(cmd1)


    #
    #函数：下载依赖包
    #
    @dbus.service.method(INTERFACE, in_signature='s', out_signature='')
    def down_dep_pkg(self,appname):
        cmd = "apt-get --download-only install "+appname+" -y"
        os.system(cmd)
        self.down_dep_finish(appname)

    #
    #信号：确认依赖包下载结束
    #
    @dbus.service.signal(INTERFACE, signature="s")
    def down_dep_finish(self,appname):
        pass


